from torch import nn
import torch
import math
import numpy as np

from torch.nn.init import xavier_normal_, xavier_uniform_, constant_
import warnings
from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class ErrorTracking:
    def __init__(self, item_prob, effective_noise_multiplier, config):
        self.d_model = config['d_model']
        self.is_vanilla = config['model'] == 'vanilla'
        self.norm = torch.distributions.normal.Normal(0, 1)

        self.error_W = effective_noise_multiplier ** 2
        self.error_E = (effective_noise_multiplier / (item_prob + 1e-6)) ** 2

        if self.is_vanilla == True:
            self.error_W = 0
            self.error_E = torch.zeros_like(self.error_E)
    
    def CDF(self, x):
        return self.norm.cdf(x)
    
    def PDF(self, x):
        return self.norm.log_prob(x).exp()

    def set_input_token(self, x):
        self.s0 = self.error_E[x]
    
    def get_token_error(self, x):
        return self.error_E[x]

    def linear_propagation(self, x, w, x_s=None, w_s=None):
        if x_s == None:
            x_s = self.s0
        if w_s == None:
            w_s = self.error_W
        return (x_s * w_s + x_s * w**2 + w_s * x**2) * self.d_model

    def nonlinear_propagation(self, mu, sigma):
        nu = sigma
        gamma = mu / sigma

        new_mu = mu * self.CDF(gamma) + nu * self.PDF(gamma)
        new_mu2 = (mu**2 + sigma**2) * self.CDF(gamma) + mu * nu * self.PDF(gamma)
        return new_mu2 - new_mu**2


def softmax(x, s, normalize=True):
    """
    Re-Attention Mechanism
    """
    x = (x - s).clamp_(-1)
    if normalize == False:
        return x
    return F.softmax(x, dim=-1)


class GELU(nn.Module):
    """
    Implementation of the gelu activation function.
    From OpenAI GPT's gelu 
    """
    def forward(self, x):
        # return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
        return 0.5 * x * (1.0 + torch.erf(x / math.sqrt(2.0)))


class PositionwiseFeedForward(nn.Module):
    "Implements FFN equation."
    def __init__(self, d_model, dropout=0.2):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_model)
        self.w_2 = nn.Linear(d_model, d_model)
        self.dropout = nn.Dropout(dropout)
        self.activation = nn.ReLU()

    def forward(self, x, errortracking):
        errortracking.s_o = errortracking.linear_propagation(x.mean().detach(), self.w_1.weight.mean().detach())
        x = self.w_1(x)

        errortracking.s_o = errortracking.nonlinear_propagation(x.mean().detach(), errortracking.s_o)
        x = self.dropout(self.activation(x))

        errortracking.s_o = errortracking.linear_propagation(x.mean().detach(), self.w_2.weight.mean().detach())
        x = self.dropout(self.w_2(x))

        errortracking.s0 = errortracking.s0 + errortracking.s_o
        return x


class SublayerConnection(nn.Module):
    """
    A residual connection followed by a layer norm.
    """
    def __init__(self, d_model, dropout=0.2):
        super(SublayerConnection, self).__init__()
        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        "Apply residual connection to any sublayer with the same size."
        x = self.norm(x)
        return x + self.dropout(sublayer(x))


class Attention(nn.Module):
    """
    Compute 'Scaled Dot Product Attention
    """
    def forward(self, query, key, value, errortracking, mask=None, dropout=None):
        scores = torch.matmul(query, key.transpose(-2, -1)) \
                 / math.sqrt(query.size(-1))

        if mask is not None:
            padding_mask, timeline_mask = mask
            scores = scores.masked_fill(padding_mask == 0, -1e9)
            scores = scores.masked_fill(timeline_mask == 0, -1e9)
        
        errortracking.s_att = errortracking.s_k * errortracking.s_q
        p_attn = softmax(scores, errortracking.s_att[:, None, None, :])

        if dropout is not None:
            p_attn = dropout(p_attn)

        return torch.matmul(p_attn, value), p_attn


class MultiHeadedAttention(nn.Module):
    """
    Take in model size and number of heads.
    """

    def __init__(self, d_model, num_heads, dropout=0.2):
        super().__init__()
        assert d_model % num_heads == 0

        # We assume d_v always equals d_k
        self.d_k = d_model // num_heads
        self.h = num_heads

        self.linear_layers = nn.ModuleList([nn.Linear(d_model, d_model) for _ in range(3)])
        self.output_linear = nn.Linear(d_model, d_model)
        self.attention = Attention()

        self.dropout = nn.Dropout(p=dropout)

    def forward(self, query, key, value, errortracking, mask=None):
        batch_size = query.size(0)

        # 1) Do all the linear projections in batch from d_model => h x d_k
        query, key, value = [l(x).view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
                             for l, x in zip(self.linear_layers, (query, key, value))]

        errortracking.s_q, errortracking.s_k, errortracking.s_v = \
            [errortracking.linear_propagation(x.mean().detach(), l.weight.mean().detach()) \
                 for l, x in zip(self.linear_layers, (query, key, value))]

        # 2) Apply attention on all the projected vectors in batch.
        x, attn = self.attention(query, key, value, errortracking, mask=mask, dropout=self.dropout)

        errortracking.s0 = errortracking.s0 + errortracking.s_att * errortracking.s_v

        # 3) "Concat" using a view and apply a final linear.
        x = x.transpose(1, 2).contiguous().view(batch_size, -1, self.h * self.d_k)

        return self.output_linear(x)


class PositionalEmbedding(nn.Module):
    def __init__(self, max_len, d_model):
        super().__init__()

        # Compute the positional encodings once in log space.
        self.pe = nn.Embedding(max_len, d_model)

    def forward(self, x):
        positions = torch.arange(0, x.shape[1]).repeat(x.shape[0], 1).cuda()
        return self.pe(positions)

class ItemEmbedding(nn.Module):
    def __init__(self, item_num, d_model):
        super().__init__()

        self.item_num = item_num
        self.embedding_dim = d_model

        self.item_emb = torch.nn.Embedding(self.item_num + 1, self.embedding_dim, padding_idx=0)
        self.item_emb.weight.padding = 0
    
    def forward(self, x):
        return self.item_emb(x)


class TransformerEmbedding(nn.Module):
    """
    Transformer Embedding which is consisted with under features
        1. TokenEmbedding : normal embedding matrix
        2. PositionalEmbedding : adding positional information
        sum of all these features are output of TransformerEmbedding
    """

    def __init__(self, item_num, d_model, max_len, dropout=0.2):
        """
        :param item_num: total item number
        :param d_model: embedding size of token embedding
        :param dropout: dropout rate
        """
        super().__init__()
        self.item = ItemEmbedding(item_num, d_model)
        self.position = PositionalEmbedding(max_len, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        x = self.item(x) + self.position(x)
        return self.dropout(x)


class TransformerBlock(nn.Module):
    def __init__(self, d_model, num_heads, dropout):
        """
        :param d_model: hidden size of transformer
        :param num_heads: head sizes of multi-head attention
        :param dropout: dropout rate
        """

        super().__init__()
        self.attention = MultiHeadedAttention(d_model, num_heads, dropout)
        self.feed_forward = PositionwiseFeedForward(d_model, dropout)
        self.input_sublayer = SublayerConnection(d_model, dropout)
        self.output_sublayer = SublayerConnection(d_model, dropout)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, mask, errortracking=None):
        x = self.input_sublayer(x, lambda _x: self.attention.forward(_x, _x, _x, errortracking, mask=mask))
        x = self.output_sublayer(x, lambda _x: self.feed_forward.forward(_x, errortracking))
        return x


class Transformer(nn.Module):
    def __init__(self, item_num, config):
        super().__init__()
        # fix_random_seed_as(args.model_init_seed)
        self.dataset = config['dataset']
        self.item_num = item_num
        self.d_model = config['d_model']
        num_blocks = config['num_blocks']
        num_heads = config['num_heads']
        dropout = config['dropout']
        max_len = config['max_len']

        self.embedding = TransformerEmbedding(self.item_num, self.d_model, max_len, dropout)

        # multi-layers transformer blocks, deep network
        self.transformer_blocks = nn.ModuleList(
            [TransformerBlock(self.d_model, num_heads, dropout) for _ in range(num_blocks)])

        self.ln_f = nn.LayerNorm(self.d_model)

        self.apply(self._init_weights)
    

    def _init_weights(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=0.01)
        elif isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=0.01)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def set_errortrackikng(self, errortracking):
        self.errortracking = errortracking

    def get_context(self, x):
        padding_mask = (x > 0).unsqueeze(1).repeat(1, x.size(1), 1).unsqueeze(1)        # B, 1, L, L
        timeline_mask = torch.tril(torch.ones((1, 1, x.size(1), x.size(1))).cuda())

        # embedding the indexed sequence to sequence of vectors
        x = self.embedding(x)

        # running over multiple transformer blocks
        for transformer in self.transformer_blocks:
            x = transformer.forward(x, (padding_mask, timeline_mask), self.errortracking)
        
        x = self.ln_f(x)
        return x

    def forward(self, x, yp=None, yn=None):
        self.errortracking.set_input_token(x)
        x = self.get_context(x)
        
        # get candicates' embeddings
        idx = torch.LongTensor(torch.arange(0, self.item_num + 1))
        idx = idx.expand(x.shape[0], idx.shape[0]).cuda()
        embs = self.embedding.item(idx)
        s = self.errortracking.get_token_error(idx)

        # compute logits
        logits = torch.bmm(x, embs.transpose(-1, -2))
        s = s * self.d_model / 2
        logits = softmax(logits, s[:, None, :], False)

        return logits


    def predict(self, x):
        self.errortracking.set_input_token(x)
        x = self.get_context(x)

        next_item = x[:, -1, :]

        logits = next_item @ self.embedding.item.item_emb.weight.T

        idx = torch.LongTensor(torch.arange(0, self.item_num + 1)).cuda()
        s = self.errortracking.get_token_error(idx)
        s = s * self.d_model / 2
        logits = softmax(logits, s[None, :], False)

        return logits