"""
This module is a collection of grad samplers 
- methods to calculate per sample grad norms or gradients
for a layer given 1) inputs, AND/OR 2) grad_outputs.

Supports BK (book-keeping) introduced in 
Bu et al. (arXiv:2210.00038)
Differentially Private Optimization on Large Model at Small Cost

and BiTFiT (bias-term fine-tuning) introduced in
Bu et al. (aarXiv:2210.00036)
Differentially Private Bias-Term only Fine-tuning of Foundation Models

Highlights: this code uses the important "mixed ghost norm" trick to achieve its high efficiency,
adapted and improved from 'Scalable and Efficient Training of Large Convolutional Neural Networks with Differential Privacy'
by Bu et al. See their Section 4.

A large portion of this code is adapted Opacus v0.15 (https://github.com/pytorch/opacus), 
from Private-transformers v0.2.3 (https://github.com/lxuechen/private-transformers),
and from Private-vision v0.1.0 (https://github.com/woodyx218/private_vision)
"""

import torch
from torch import nn
from torch.functional import F

def mixed_ghost_norm(layer,A,B,conv=False):
    # for linear layers, A is activation, B is backprops;
    # for conv layers, A is unfolded activation, B is inverse folded (flattened) backprops;
    if not hasattr(layer, "use_gc"): # use ghost clipping or not
        if conv==False:
            T = torch.prod(torch.Tensor(list(A.shape[1:-1]))).item()
            #assert T == torch.prod(torch.Tensor(list(B.shape[1:-1]))).item()
            d = A.shape[-1]
            p = B.shape[-1]
        else:
            T = A.shape[-1]
            assert T == B.shape[-1]
            d = A.shape[1]
            p = B.shape[1]
        d_times_p = torch.prod(torch.Tensor(list(layer.weight.size())))
        layer.use_gc = bool(2*T**2 <= d_times_p)
        #assert d*p == d_times_p
        #print(layer,'\n use ghost clip: ',layer.use_gc,'\n T= ',T,';d= ',d,';p= ',p,';2T^2= ',2*T**2,';pd= ',p*d)
        B=100
        layer.BKtime=B*(6*T*d*p+layer.use_gc*2*(T**2)*(p+d))
        layer.nonDPtime=B*(6*T*d*p)
        layer.GhostCliptime=B*(10*T*d*p+2*(T**2)*(p+d))
        layer.Opacustime=B*(8*T*d*p)

        layer.BKspace=p*d+B*(min(2*(T**2),p*d)+T*(3*d+p))
        layer.nonDPspace=p*d+B*T*(3*d+p)
        layer.GhostClipspace=p*d+(B*2*(T**2)+B*T*(3*d+p))
        layer.Opacusspace=p*d+(B*d*p+B*T*(3*d+p))

def sum_over_all_but_batch_and_last_n(tensor: torch.Tensor, n_dims: int) -> torch.Tensor:
    if tensor.dim() == n_dims + 1:
        return tensor
    else:
        dims = list(range(1, tensor.dim() - n_dims))
        return tensor.sum(dim=dims)


def _light_linear_weight_norm_sample(A, B) -> torch.Tensor:
    """Compute gradient sample norm for the weight matrix in a linear layer."""
    if A.dim() == 2 and B.dim() == 2:
        return _light_linear_weight_norm_sample_non_sequential(A, B)
    elif A.dim() == 3 and B.dim() == 3:
        return _light_linear_weight_norm_sample_sequential(A, B)
    else:
        raise ValueError(f"Unexpected input shape: {A.size()}, grad_output shape: {B.size()}")


@torch.jit.script
def _light_linear_weight_norm_sample_sequential(A, B):
    """Lightweight norm computation in ghost clipping.

    Linear algebra identity trick -- Eq. 3 in the paper.
    """
    #return torch.sqrt((torch.einsum('bTd,bSd->bTS',A,A)*torch.einsum('bTp,bSp->bTS',B,B)).sum(dim=(1, 2)))
    return torch.sqrt((torch.bmm(A, A.transpose(-1, -2)) * torch.bmm(B, B.transpose(-1, -2))).sum(dim=(1, 2)))


@torch.jit.script
def _light_linear_weight_norm_sample_non_sequential(A, B):
    """The Goodfellow trick, i.e., Frobenius norm equal to product of 2-norms."""
    return A.norm(2, dim=1) * B.norm(2, dim=1)

@torch.jit.script
def _light_linear_bias_norm_sample(B):
    if B.dim() == 2:
        return B.norm(2, dim=1)
    elif B.dim() == 3:
        return B.sum(dim=1).norm(2, dim=1)
    else:
        raise ValueError(f"Unexpected grad_output shape: {B.size()}")


def _create_or_extend_norm_sample(param: torch.Tensor, norm_sample: torch.Tensor) -> None:
    """Creates a ``norm_sample`` attribute in the given parameter."""
    if not hasattr(param, "norm_sample") or param.norm_sample == None:
        param.norm_sample = norm_sample
    else:
        """
        Phantom Clipping
        """
        a1 = param.a1
        g1 = param.g1
        g2 = param.g2
        param.norm_sample = torch.sqrt(param.norm_sample ** 2 + norm_sample ** 2 + 2 * (g1 * g2.gather(1, a1[:, :, None].repeat(1, 1, g2.shape[-1]))).sum((1, 2)))

        del param.a1, param.g1, param.g2


def _create_or_extend_grad_sample(param: torch.Tensor, grad_sample: torch.Tensor) -> None:
    """Creates a ``grad_sample`` attribute in the given parameter or accumulate the existing tensor."""
    assert grad_sample.shape[1:] == param.shape, f"grad_sample.size()={grad_sample.size()}, param.size()={param.size()}"
    if hasattr(param, "grad_sample"):
        param.grad_sample += grad_sample.detach()
    else:
        param.grad_sample = grad_sample.detach()


def _compute_linear_grad_sample(layer: nn.Linear, A: torch.Tensor, B: torch.Tensor, clipping_mode: str) -> None:
    """Computes per sample gradients for `nn.Linear` layer.
    A is activations or layer's input, see autograd_grad_sample line 229; B is output gradient
    This function is written in an unusually bespoke way to avoid using `torch.einsum`.
    """
    if A!=None:
        if clipping_mode in ['MixGhostClip','MixOpt']:
            mixed_ghost_norm(layer, A, B)
        else:
            layer.use_gc=True
        
        
        if A.dim()>3:
            A=torch.flatten(A,start_dim=1,end_dim=-2)
            B=torch.flatten(B,start_dim=1,end_dim=-2)
          

        if layer.use_gc==True:
            #--- compute weight gradient norm
            _create_or_extend_norm_sample(layer.weight, _light_linear_weight_norm_sample(A, B))
        else:
            ## Or use Line 105 (v0.1.0) https://github.com/lxuechen/private-transformers/blob/main/private_transformers/privacy_utils/supported_layers_grad_samplers.py
            layer.weight.grad_sample = torch.einsum('b...d, b...p-> bpd', A, B)

            _create_or_extend_norm_sample(layer.weight, torch.sqrt(torch.sum(layer.weight.grad_sample**2, dim=(1, 2))))
            if clipping_mode!='MixOpt':
                del layer.weight.grad_sample
    
    #--- compute bias gradient norm
    if layer.bias is not None:
        _create_or_extend_norm_sample(layer.bias, _light_linear_bias_norm_sample(B))
        if B.dim() == 3:
            grad_bias = B.sum(dim=1)
        elif B.dim() == 2:
            grad_bias = B
        _create_or_extend_grad_sample(layer.bias, grad_bias)        



def _compute_layer_norm_grad_sample(
    layer: nn.LayerNorm,
    A: torch.Tensor, B: torch.Tensor, 
    clipping_mode: str) -> None:
    """Computes per sample gradients for normalization layers."""
    if A!=None:
        #--- weight, compute gradient norm
        grad_sample = sum_over_all_but_batch_and_last_n(
            F.layer_norm(A, layer.normalized_shape, eps=layer.eps) * B,
            layer.weight.dim(),
        )
        norm_sample = grad_sample.flatten(start_dim=1).norm(2, dim=1)
        _create_or_extend_norm_sample(layer.weight, norm_sample)
    
    #--- bias, compute gradient norm
    if layer.bias is not None:
        grad_sample = sum_over_all_but_batch_and_last_n(B, layer.bias.dim())        
        _create_or_extend_norm_sample(layer.bias, grad_sample.flatten(start_dim=1).norm(2, dim=1))
        _create_or_extend_grad_sample(layer.bias, grad_sample)


def _compute_embedding_grad_sample(layer: nn.Embedding, A: torch.Tensor, B: torch.Tensor, clipping_mode: str) -> None:
    """Computes per sample gradients for `nn.Embedding` layer."""
    if clipping_mode in ['MixGhostClip','MixOpt']:
        mixed_ghost_norm(layer, A, B)
    else:
        layer.use_gc=True

    padding_idx = layer.padding_idx
    norm_sample = None
    if padding_idx is not None:
        if type(layer.activations) == tuple:
            if not hasattr(layer.weight, "g2") or layer.weight.g2 == None:
                B[:, padding_idx] = 0
                layer.weight.g2 = B
            else:
                not_AAt: torch.Tensor = ~A[:, :, None].eq(A[:, None, :])
                not_AAt.bitwise_or_((A[:, :, None] == padding_idx) | (A[:, None, :] == padding_idx))
                norm_sample = torch.sqrt((torch.bmm(B, B.transpose(-1, -2)).masked_fill(not_AAt, 0)).sum(dim=(1, 2)))

                layer.weight.g1 = B
                layer.weight.a1 = A
        
        else:
            if A.shape[1] == layer.weight.shape[0]:
                B[:, padding_idx] = 0

    if norm_sample == None:
        norm_sample = B.norm(2, dim=(1, 2))
    _create_or_extend_norm_sample(layer.weight, norm_sample)
    

#% compute clipped weight gradient    
def _clip_linear_grad(layer: nn.Linear, A: torch.Tensor, B: torch.Tensor, C) -> None:
    try:
        grad_weight = torch.einsum('b...,b->...',layer.weight.grad_sample,C)
        del layer.weight.grad_sample
    except:
        grad_weight = torch.einsum('...p,...d->pd',B,A)
    return grad_weight



def _clip_layer_norm_grad(layer: nn.LayerNorm, A: torch.Tensor, B: torch.Tensor, C) -> None:
    def sum_over_all_but_batch_and_last_n(tensor: torch.Tensor, n_dims: int) -> torch.Tensor:
        if tensor.dim() == n_dims + 1:
            return tensor
        else:
            dims = list(range(1, tensor.dim() - n_dims))
            return tensor.sum(dim=dims)
    grad_weight = sum_over_all_but_batch_and_last_n(
        F.layer_norm(A, layer.normalized_shape, eps=layer.eps) * B,
        layer.weight.dim(),).sum(dim=0)
    return grad_weight


def _clip_instance_norm_grad(layer, A: torch.Tensor, B: torch.Tensor, C) -> None:
    grad_weight = torch.einsum('bi...->i',F.instance_norm(A, eps=layer.eps) * B)
    return grad_weight

def _clip_embedding_grad(layer: nn.Embedding, A: torch.Tensor, B: torch.Tensor, C) -> None:
    if hasattr(layer.weight, 'current_item_num'):
        if A.shape[1] == layer.weight.current_item_num + 1 or A.shape[1] == layer.weight.shape[0]:
            grad_weight = B.sum(0)

        else:
            # A = F.one_hot(A, num_classes=layer.weight.current_item_num + 1).to(B)  # (batch_size, seq_len, vocab_dim,)
            A = F.one_hot(A, num_classes=layer.weight.shape[0]).to(B)  # (batch_size, seq_len, vocab_dim,)
            grad_weight = torch.einsum('...p,...d->dp',B,A)
        ## `torch.nn.Embedding` layers don't accumulate gradient on the padding_idx position.
        ##   We do the same for `grad_sample`.
        if layer.padding_idx is not None:
            # `grad_sample` has size (batch_size, num_vocab, embedding_dim).
            grad_weight[layer.padding_idx, :] = 0.
        
        nonce = torch.zeros(layer.weight.shape[0] - grad_weight.shape[0], layer.weight.shape[1]).cuda()
        grad_weight = torch.cat([grad_weight, nonce], 0)

        
    else:
        if A.shape[1] == layer.weight.shape[0]:
            grad_weight = B.sum(0)
        else:
            A = F.one_hot(A, num_classes=layer.weight.shape[0]).to(B)  # (batch_size, seq_len, vocab_dim,)
            grad_weight = torch.einsum('...p,...d->dp',B,A)

        if layer.padding_idx is not None:
            # `grad_sample` has size (batch_size, num_vocab, embedding_dim).
            grad_weight[layer.padding_idx, :] = 0.

    return grad_weight


_supported_layers_norm_sample_AND_clipping = {
    nn.Embedding: (_compute_embedding_grad_sample, _clip_embedding_grad),
    nn.Linear: (_compute_linear_grad_sample, _clip_linear_grad),
    nn.LayerNorm: (_compute_layer_norm_grad_sample, _clip_layer_norm_grad),
}

#%  we need param.summed_clipped_grad to avoid contamination from non-private .grad
def _create_or_extend_summed_clipped_grad(param: torch.Tensor, summed_clipped_grad: torch.Tensor) -> None:
    """Adds summed clipped gradient (not per-sample) to param.summed_clipped_grad or accumulate the existing tensor."""
    if summed_clipped_grad.shape != param.shape:
        summed_clipped_grad = summed_clipped_grad.T
    assert summed_clipped_grad.shape == param.shape, f"summed clipped grad.size()={summed_clipped_grad.size()}, param.size()={param.size()}"

    if hasattr(param, "summed_clipped_grad") and param.summed_clipped_grad!=None:
        param.summed_clipped_grad += summed_clipped_grad.detach()
    else:
        param.summed_clipped_grad = summed_clipped_grad.detach()
