import pandas as pd
import numpy as np
from sklearn import preprocessing
import argparse

dataset = 'ml1m'

raw_sep_dict = {
    'ml1m': "::"
}

raw_file_dict = {
    'ml1m': "ml1m/ratings.dat"
}

raw_sep = raw_sep_dict[dataset]
raw_file = raw_file_dict[dataset]

filter_min = 5

processed_file_prefix = f"{dataset}/"

# ================================================================================
# obtain implicit feedbacks
df = pd.read_csv(raw_file, sep=raw_sep, header=None, names=['user_id','item_id','rating','timestamp'])
df = df[['user_id','item_id','timestamp']]

print("==== statistic of raw data ====")
print("#users: %d" % len(df.user_id.unique()))
print("#items: %d" % len(df.item_id.unique()))
print("#actions: %d" % len(df))

# ========================================================
# sort by time
df.sort_values(by=['timestamp'], kind='mergesort', ascending=True, inplace=True)

# ========================================================
# drop duplicated user-item pairs
df.drop_duplicates(subset=['user_id','item_id'], keep='first', inplace=True)

# ========================================================
# discard cold-start items
count_i = df.groupby('item_id').user_id.count()
item_keep = count_i[count_i >= filter_min].index
df = df[df['item_id'].isin(item_keep)]

# discard cold-start users
count_u = df.groupby('user_id').item_id.count()
user_keep = count_u[count_u >= filter_min].index
df = df[df['user_id'].isin(user_keep)]

# renumber user ids and item ids
le = preprocessing.LabelEncoder()
df['user_id'] = le.fit_transform(df['user_id'])
df['item_id'] = le.fit_transform(df['item_id']) + 1

# output statistical information
print("==== statistic of processed data (whole) ====")
n = len(df.user_id.unique())
m = len(df.item_id.unique())
p = len(df)
print("#users: %d" % n)
print("#items: %d" % m)
print("#actions: %d" % p)
print("density: %.4f" % (p/n/m))

count_u = df.groupby(['user_id']).item_id.count()
print("min #actions per user: %.2f" % count_u.min())
print("max #actions per user: %.2f" % count_u.max())
print("ave #actions per user: %.2f" % count_u.mean())

# ========================================================
# split data into test set, valid set and train set, 
# adopting the leave-one-out evaluation for next-item recommendation task

# ========================================
# obtain possible records in test set
df_test = df.groupby(['user_id']).tail(1)

df.drop(df_test.index, axis='index', inplace=True)

# ========================================
# obtain possible records in valid set
df_valid = df.groupby(['user_id']).tail(1)

df.drop(df_valid.index, axis='index', inplace=True)

# ========================================
# drop cold-start items in valid set and test set
df_valid = df_valid[df_valid.item_id.isin(df.item_id)]
df_test = df_test[df_test.user_id.isin(df_valid.user_id) & (
    df_test.item_id.isin(df.item_id) | df_test.item_id.isin(df_valid.item_id))]

# output data file
df_valid.to_csv(processed_file_prefix + f"{dataset}_valid.csv", header=False, index=False)
df_test.to_csv(processed_file_prefix + f"{dataset}_test.csv", header=False, index=False)
df.to_csv(processed_file_prefix + f"{dataset}_train.csv", header=False, index=False)

# output statistical information
print("==== statistic of processed data (split) ====")
print("#train_users: %d" % len(df.user_id.unique()))
print("#train_items: %d" % len(df.item_id.unique()))
print("#valid_users: %d" % len(df_valid.user_id.unique()))
print("#test_users: %d" % len(df_test.user_id.unique()))