import torch
import torch.backends.cudnn as cudnn
import torch.optim
import torch.distributed as dist
from d2l import torch as d2l
import numpy as np

import numpy as np
import bottleneck as bn


def NDCG_binary_at_k_batch(X_pred, heldout_batch, k=5):
    '''
    normalized discounted cumulative gain@k for binary relevance
    ASSUMPTIONS: all the 0's in heldout_data indicate 0 relevance
    '''
    batch_users = X_pred.shape[0]
    idx_topk_part = bn.argpartition(-X_pred, k, axis=1)
    topk_part = X_pred[np.arange(batch_users)[:, np.newaxis], idx_topk_part[:, :k]]
    idx_part = np.argsort(-topk_part, axis=1)
    # topk predicted score
    idx_topk = idx_topk_part[np.arange(batch_users)[:, np.newaxis], idx_part]
    # build the discount template
    tp = 1. / np.log2(np.arange(2, k + 2))
    DCG = (heldout_batch[np.arange(batch_users)[:, np.newaxis], idx_topk] * tp).sum(axis=1)
    IDCG = np.array([(tp[:min(n, k)]).sum() for n in heldout_batch.sum(axis=1).astype(np.int32)])
    return DCG / IDCG


def Recall_Precision_F1_OneCall_at_k_batch(X_pred, heldout_batch, k=5):
    batch_users = X_pred.shape[0]
    idx = bn.argpartition(-X_pred, k, axis=1)
    X_pred_binary = np.zeros_like(X_pred, dtype=bool)
    X_pred_binary[np.arange(batch_users)[:, np.newaxis], idx[:, :k]] = True
    X_true_binary = (heldout_batch > 0)
    tmp = (np.logical_and(X_true_binary, X_pred_binary).sum(axis=1)).astype(np.float32)
    recall = tmp / X_true_binary.sum(axis=1)
    precision = tmp / k
    f1 = 2 * recall * precision / (recall + precision)
    oneCall = (tmp > 0).astype(np.float32)
    return recall, precision, f1, oneCall


def checknan(a):
    a[np.isnan(a)] = 0
    return a.mean()


@torch.no_grad()
def eval(model, dataset, data_loader, topk=10):
    model.eval()
    NDCG10 = d2l.AverageMeter()
    HT10 = d2l.AverageMeter()


    for i, (X, train_items, yp) in enumerate(data_loader):
        X = X.cuda(non_blocking=True)
        train_items = train_items.cuda(non_blocking=True)
        yp = yp.cuda(non_blocking=True)

        if hasattr(model, 'predict') and callable(model.predict):
            predictions = model.predict(X)
        else:
            predictions = model._module.predict(X)
        
        predictions = predictions.scatter_(1, train_items, -torch.inf)

        yp = torch.zeros_like(predictions).scatter_(1, yp, 1)

        predictions = predictions[:, 1:].cpu().numpy()
        yp = yp[:, 1:].cpu().numpy()

        n_10 = NDCG_binary_at_k_batch(predictions, yp, topk)
        r_10, p_10, f_10, o_10 = Recall_Precision_F1_OneCall_at_k_batch(predictions, yp, topk)

        NDCG10.update(checknan(n_10), X.shape[0])
        HT10.update(checknan(r_10), X.shape[0])

   
    model.train()
    return NDCG10.avg, HT10.avg
