import argparse
import os
import random
import logging
import numpy as np
import time

import torch
import torch.backends.cudnn as cudnn
import torch.optim
from d2l import torch as d2l
from tensorboardX import SummaryWriter
from torch import nn
import copy
import yaml
import torch.nn.functional as F

from PhantomClipping import PrivacyEngine
from DPFormer import *
from mydatasets import *

from evaluation import *

local_time = time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime())

parser = argparse.ArgumentParser()

parser.add_argument('--user', default='root', type=str)
parser.add_argument('--date', default=local_time, type=str)
parser.add_argument('--config', default='DPFormer_ml1m', type=str)
# parser.add_argument('--config', default='Transformer_ml1m', type=str)
# parser.add_argument('--config', default='Transformer_Amazon', type=str)
# parser.add_argument('--config', default='DPFormer_Amazon', type=str)
parser.add_argument('--num_workers', default=4, type=int)

parser.add_argument('--resume', default=False, type=bool)
args = parser.parse_args()

config = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configs', f'{args.config}.yaml'), encoding="UTF-8")
config = yaml.load(config, Loader=yaml.FullLoader)

d2l.init_Seed(config['seed'])

# logging
log_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'log', f"{args.config}-{args.date[8:]}")
log_file = log_dir + '.txt'
d2l.log_args(log_file)
logging.info('--------------------------------------This is all argsurations----------------------------------')
for arg in vars(args):
    logging.info('{}={}'.format(arg, getattr(args, arg)))
for arg in config:
    logging.info('{}={}'.format(arg, config[arg]))
logging.info('----------------------------------------This is a halvingline----------------------------------')

# checkpoints
checkpoint_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'checkpoint', f"{args.config}-{args.date[8:]}")
if not os.path.exists(checkpoint_dir):
    os.makedirs(checkpoint_dir)

# tensorboard
writer = SummaryWriter(f"runs/{args.config}-{args.date[8:]}")

dataset = getDataset(config)

num_users = dataset.train.user_id.max()
num_items = dataset.train.item_id.max()

train_loader = torch.utils.data.DataLoader(dataset=TrainDataset(dataset, config), batch_size=config['MAX_PHYSICAL_BATCH_SIZE'],
                            drop_last=True, num_workers=args.num_workers, pin_memory=True, shuffle=True)
test_loader = torch.utils.data.DataLoader(dataset=TestDataset(dataset, config), batch_size=config['batch_size'],
                            drop_last=False, num_workers=args.num_workers, pin_memory=True, shuffle=False)

loss_fn = nn.CrossEntropyLoss()

# prepare model
model = Transformer(num_items, config).cuda()

for name, param in model.named_parameters():
    print(name)
    print(param.shape, param.device)

optimizer = torch.optim.Adam(model.parameters(), lr=config['lr'], weight_decay=config['wd'])

# ======DP============================
privacy_engine = PrivacyEngine(
    model,
    batch_size=config['batch_size'],
    sample_size=dataset.user_maxid,
    epochs=config['epochs'],
    target_epsilon=config['epsilon'],
    target_delta=config['delta'],
    clipping_fn='automatic',
    clipping_mode='MixOpt',
    origin_params=None,
    clipping_style='all-layer',
)
privacy_engine.attach(optimizer)

print('Using noise: ', privacy_engine.noise_multiplier)

errortracking = ErrorTracking(dataset.item_prob, privacy_engine.effective_noise_multiplier, config)
model.set_errortrackikng(errortracking)

"""
Training
"""
ndcg10, hit10 = eval(model, dataset, test_loader)
logging.info('Epoch: -1: NDCG@10: {:.4f}, HIT@10: {:.4f}'.format(ndcg10, hit10))

model.train()

start_time = time.time()

for epoch in range(config['epochs']):
    start_epoch = time.time()

    losses = d2l.AverageMeter()

    def get_lr(epoch, max_epoch, init_lr, power=0.9):
        stage = max_epoch * 0.2
        if epoch + 1 <= stage:
            return init_lr * ((epoch + 1) / stage)
        
        epoch -= stage
        return round(init_lr * np.power(1-(epoch) / max_epoch, power), 8)

    d2l.adjust_learning_rate(optimizer, get_lr(epoch, config['epochs'], config['lr']))

    def train(i, batch):
        X, yp, yn, users = batch

        X = X.cuda(non_blocking=True)
        yp = yp.cuda(non_blocking=True)
        yn = yn.cuda(non_blocking=True)

        logits = model(X, yp, yn)

        loss_fn = nn.CrossEntropyLoss(ignore_index=0)
        loss = loss_fn(logits.reshape(-1, logits.shape[-1]), yp.reshape(-1))

        losses.update(loss.data.item(), X.shape[0])
        loss.backward()
        
        if (i+1) % (config['batch_size'] // config['MAX_PHYSICAL_BATCH_SIZE']) == 0:
            optimizer.step()
            optimizer.zero_grad()


    for i, batch in enumerate(train_loader):
        train(i, batch)

    writer.add_scalar(f'loss', losses.avg, epoch)

    if (epoch + 1) % 5 == 0:
        ndcg10, hit10 = eval(model, dataset, test_loader)
        writer.add_scalar(f'ndcg10', ndcg10, epoch)
        writer.add_scalar(f'hit10', hit10, epoch)
    
    logging.info('Epoch: {}: loss: {:.4f}, NDCG@10: {:.4f}, HIT@10: {:.4f}'.format(epoch, losses.avg, ndcg10, hit10))

    end_epoch = time.time()
    epoch_time_minute = (end_epoch-start_epoch)/60
    logging.info('Current epoch time consumption: {:.2f} minutes!'.format(epoch_time_minute))

writer.close()

final_name = os.path.join(checkpoint_dir, 'model_epoch_last.pth')
torch.save({
    'state_dict': model.state_dict(),
}, final_name)

end_time = time.time()
total_time = (end_time-start_time)/3600
logging.info('The total training time is {:.2f} hours'.format(total_time))
logging.info('----------------------------------The training process finished!-----------------------------------')