import torch
from torch.functional import split
import torch.nn as nn
from torch.autograd import Variable
import torch.backends.cudnn as cudnn
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.data import DataLoader, Dataset
import numpy as np
import os
import pandas as pd
import random
import logging
import time
import copy
from scipy import sparse
import pickle


def getDataset(config):
    file_name = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', config['dataset'], 'pre_load')
    if not os.path.exists(file_name):
        dataset = SequentialDataset(config)
        fw = open(file_name,'wb')
        pickle.dump(dataset, fw)
        fw.close()
        return dataset
    
    fr = open(file_name,'rb')
    dataset = pickle.load(fr)
    fr.close()
    return dataset


class SequentialDataset(Dataset):
    def __init__(self, config):
        self.dataset = config['dataset']
        self.maxlen = config['max_len']
        
        self.model = config['model']

        logging.info('Loading {} ...'.format(self.dataset))
        start_time = time.time()

        self.load_files(self.dataset)
    
        self.train = pd.concat([self.train, self.valid], axis='index')

        # preprocessing
        self.train.sort_values(by=['user_id','timestamp'], axis='index', ascending=True, inplace=True, kind='mergesort')

        self.train = self.train.groupby(['user_id']).tail(self.maxlen + 1)

        self.user_set = set(self.train.user_id.unique())
        self.item_set = set(self.train.item_id.unique())
        self.user_maxid = np.max(self.train.user_id.unique())
        self.item_maxid = np.max(self.train.item_id.unique())


        self.train_seq = {}
        self.neg_cand = {}
        self.test_seq = []

        for u in self.user_set:
            items = self.train[self.train['user_id'] == u].item_id.values
            self.train_seq[u] = items

        start_time2 = time.time()
        logging.info('Finish Loading Training data({:.2f} min)'.format((start_time2 - start_time) / 60))

        test_user_list = list(self.test.user_id.unique())
        for u in test_user_list:
            item = self.test[self.test['user_id'] == u].item_id.values[0]
            self.test_seq.append([u, item])

        item_sorted_data = self.train.sort_values(by='item_id', axis='index', ascending=True, inplace=False)
        item_count_pd = item_sorted_data.groupby('item_id').count()
        item_count = np.zeros([self.item_maxid + 1], dtype=np.float32)
        item_count[item_count_pd.index.to_numpy()] = item_count_pd.user_id.to_numpy()
        self.item_prob = torch.tensor(item_count / self.user_maxid).cuda()
        self.item_prob = self.item_prob / self.item_prob.max() * 0.5

        logging.info('Finish reading ({:.2f} min)'.format((time.time() - start_time2) / 60))


    def load_files(self, dataset):
        file_prefix = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', dataset)
        self.train = pd.read_csv(os.path.join(file_prefix, f'{dataset}_train.csv'), header=None, 
                                    names=['user_id', 'item_id', 'timestamp'], 
                                    dtype={'user_id': np.int32, 'item_id': np.int32, 'timestamp': np.int32})
        self.valid = pd.read_csv(os.path.join(file_prefix, f'{dataset}_valid.csv'), header=None, 
                                    names=['user_id', 'item_id', 'timestamp'], 
                                    dtype={'user_id': np.int32, 'item_id': np.int32, 'timestamp': np.int32})
        self.test = pd.read_csv(os.path.join(file_prefix, f'{dataset}_test.csv'), header=None, 
                                    names=['user_id', 'item_id', 'timestamp'], 
                                    dtype={'user_id': np.int32, 'item_id': np.int32, 'timestamp': np.int32})


class TrainDataset(Dataset):
    def __init__(self, dataset, config):
        self.core = dataset

    def convert(self, seq):
        seq = np.array(seq)
        xlen = (seq == 0).sum()
        return np.pad(seq[xlen:], (0, xlen), 'constant')


    def __getitem__(self, u):
        train_seq = self.core.train_seq[u]

        if len(train_seq) >= self.core.maxlen+1:
            train_seq = train_seq[-(self.core.maxlen + 1):]

        train_seq = np.pad(train_seq, (self.core.maxlen+1-len(train_seq), 0), 'constant')

        x = train_seq[:-1]
        yp = train_seq[1:]

        yn = np.sort(np.array(list(self.core.item_set.difference(set(self.core.train_seq[u])))))[:1000]

        return torch.LongTensor(x), torch.LongTensor(yp), torch.LongTensor(yn), torch.LongTensor([u])


    def __len__(self):
        return len(self.core.user_set)


class TestDataset(Dataset):
    def __init__(self, dataset, config):
        self.core = dataset

    def convert(self, seq):
        seq = np.array(seq)
        xlen = (seq == 0).sum()
        return np.pad(seq[xlen:], (0, xlen), 'constant')


    def __getitem__(self, u):
        u, yp = self.core.test_seq[u]

        train_seq = self.core.train_seq[u]
        train_seq = np.pad(train_seq, (self.core.maxlen+1-len(train_seq), 0), 'constant')

        x = train_seq[1:]

        train_seq = self.core.train_seq[u]
        train_seq = np.pad(train_seq, (self.core.item_maxid-len(train_seq), 0), 'constant')

        return torch.LongTensor(x), torch.LongTensor(train_seq), torch.LongTensor([yp])


    def __len__(self):
        return len(self.core.test_seq)